"""API client for FusionSolar Kiosk."""
import asyncio
import codecs
import os
import time
from dataclasses import dataclass
from datetime import datetime

import aiohttp
from Cryptodome.Cipher import PKCS1_OAEP
from Cryptodome.Hash import SHA384
from Cryptodome.PublicKey import RSA


@dataclass
class SignalSet:
    signals: list

    def __repr__(self):
        return repr(self.signals)

    def get_ids(self):
        return [signal.id for signal in self.signals]

    def get_name_by_id(self, id):
        for signal in self.signals:
            if signal.id == int(id):
                return signal.name.lower().replace(" ", "_")


@dataclass
class Signal:
    id: int
    name: str
    unit: str

    def __repr__(self):
        return f"{self.name} ({self.unit})"


class FusionSolarRestApi:
    ENERGY_FLOW_PATH = (
        "/rest/pvms/web/station/v1/overview/energy-flow?"
        + "stationDn=%STATION%&_=%CURRENT_UTC_TIME%"
    )
    STATION_DETAIL_PATH = (
        "/rest/pvms/web/station/v1/overview/station-detail?"
        + "stationDn=%STATION%&_=%CURRENT_UTC_TIME%"
    )
    STATION_REAL_KPI = (
        "/rest/pvms/web/station/v1/overview/station-real-kpi?"
        + "stationDn=%STATION%&clientTime=%CURRENT_UTC_TIME%&"
        + "timeZone=1&_=%CURRENT_UTC_TIME%"
    )
    ENERGY_BALANCE = (
        "/rest/pvms/web/station/"
        + "v1/overview/energy-balance?stationDn=%STATION%&"
        + "timeDim=2&queryTime=%CURRENT_UTC_TIME%&timeZone=2&"
        + "timeZoneStr=Europe%2FBerlin&_=%CURRENT_UTC_TIME%"
    )
    DEVICE_STATISTICS_SIGNALS = (
        "/rest/pvms/web/device/v1/device-statistics-signal?"
        + "deviceDn=%INVERTER_STATION%"
    )
    DEVICE_SIGNALS = "/rest/pvms/web/device/v1/device-history-data"

    def __init__(self, logger, username, password, region="eu5"):
        self.logger = logger
        self._region = region
        self._username = username
        self._password = password
        self._stationdetail = None
        self._stationrealkpi = None
        self._energy_balance = None
        self._device_signals = None
        self._inverter_station = None
        self._inverter_output_power = 0
        self._from_grid = 0
        self._to_grid = 0
        self._electrical_load = 0
        self._battery_soc = 0
        self._battery_battery_power = 0
        self._battery_charge_capacity = 0
        self._battery_discharge_capacity = 0
        self._string_output_power = 0
        self._available_signals = None

    async def get_pubkey(self, session):
        # pubkey to get pubkey
        url = "https://" + self._region + ".fusionsolar.huawei.com/unisso/pubkey"
        async with session.get(url) as resp:
            self.logger.debug(f"response from {url}: {resp}")
            j = await resp.json()
            return j

    async def validate_user(self, session):
        url = (
            "https://"
            + self._region
            + ".fusionsolar.huawei.com"
            + "/unisso/"
            + self._validate_user_ver
            + "/validateUser.action"
        )

        body = {
            "multiRegionName": "",
            "organizationName": "",
            "username": self._username,
            "password": self._password_encrypted
            if self._validate_user_ver == "v3"
            else self._password,
        }

        headers = {
            "accept": "application/json, text/javascript, */*; q=0.01",
            "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
            "content-type": "application/json",
            "sec-ch-ua": '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Chrome OS"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "x-requested-with": "XMLHttpRequest",
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
        }

        async with session.post(
            url,
            json=body,
            headers=headers,
            max_redirects=20,
            params={
                "service": "/unisess/v1/auth?service=%2Fnetecowebext%2Fhome%2Findex.html",
                "decision": 1,
                "actionErrors": 465,
                "timeStamp": self._timestamp,
                "nonce": codecs.encode(os.urandom(16), "hex").decode(),
            },
        ) as resp:
            self.logger.debug(f"response from {url}: {resp}")
            j = await resp.json()
            return j

    async def login_redirect(self, session, redurl):
        headers = {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
            "Connection": "keep-alive",
            "sec-ch-ua": '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Chrome OS"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "same-origin",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
        }
        if redurl[0:4] == "http":
            self.api_base = redurl
            url = redurl
        else:
            self.api_base = None
            url = (
                "https://"
                + self._region
                + ".fusionsolar.huawei.com"
                + redurl
                + "&redirectionAddress=%2Funiportal%2Fpvmswebsite%2Fassets%2Fbuild%2Fcloud.html%23%2Fhome%2Flist"
            )
        async with session.get(url, max_redirects=20, headers=headers) as resp:
            if self.api_base is None:
                self.api_base = "https://" + resp.host
            return resp

    async def get_jsessionid(self, session):
        url = "https://" + self._region + ".fusionsolar.huawei.com/"
        async with session.get(url, max_redirects=20) as resp:
            return resp

    async def login(self):
        try:
            async with aiohttp.ClientSession(trust_env=True) as session:
                resp = await self.get_jsessionid(session)

                fusion_pubkey_json = await self.get_pubkey(session)
                self._validate_user_ver = "v2"
                if fusion_pubkey_json["enableEncrypt"]:
                    self._validate_user_ver = "v3"
                    self.public_key = RSA.import_key(
                        fusion_pubkey_json["pubKey"].encode("UTF-8")
                    )
                    cipher = PKCS1_OAEP.new(self.public_key, SHA384)
                    hex_pwd = cipher.encrypt(self._password.encode("UTF-8"))
                    self._password_encrypted = (
                        codecs.encode(hex_pwd, "base64").decode().replace("\n", "")
                        + fusion_pubkey_json["version"]
                    )
                    self._timestamp = fusion_pubkey_json["timeStamp"]

                # validateUser.action to get bspsessionid
                json_resp = await self.validate_user(session)
                if json_resp["errorCode"] == "470":
                    redurl = json_resp["respMultiRegionName"][1]
                else:
                    redurl = json_resp["redirectURL"]

                if redurl:
                    resp = await self.login_redirect(session, redurl)
                    self._cookies = session.cookie_jar

                headers = {
                    "accept": "application/json",
                    "accept-language": "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5",
                    "sec-ch-ua": '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
                    "sec-ch-ua-mobile": "?0",
                    "sec-ch-ua-platform": '"Chrome OS"',
                    "sec-fetch-dest": "empty",
                    "sec-fetch-mode": "cors",
                    "sec-fetch-site": "same-origin",
                    "x-requested-with": "XMLHttpRequest",
                    "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
                }

                if json_resp["errorCode"] == "470":
                    url = self.api_base + "/rest/dpcloud/auth/v1/keep-alive"
                    tokenid = "payload"
                else:
                    url = self.api_base + "/unisess/v1/auth/session"
                    tokenid = "csrfToken"
                async with session.get(url, headers=headers) as resp3:
                    self.logger.debug(f"response from {url}: {resp3}")
                    response = await resp3.json()
                    self._csrftoken = response[tokenid]

                headers = {
                    "accept": "application/json, text/javascript, */*; q=0.01",
                    "accept-language": "en-AT,en;q=0.9",
                    "content-type": "application/json",
                    "roarand": self._csrftoken,
                    "sec-ch-ua": '"Chromium";v="104", " Not A;Brand";v="99", "Google Chrome";v="104"',
                    "sec-ch-ua-mobile": "?0",
                    "sec-ch-ua-platform": '"Chrome OS"',
                    "sec-fetch-dest": "empty",
                    "sec-fetch-mode": "cors",
                    "sec-fetch-site": "same-origin",
                    "x-non-renewal-session": "true",
                    "x-requested-with": "XMLHttpRequest",
                    "x-timezone-offset": "120",
                    "Referer": "https://"
                    + self._region
                    + ".fusionsolar.huawei.com/pvmswebsite/assets/build/index.html",
                    "Referrer-Policy": "strict-origin-when-cross-origin",
                }
                body = {
                    "curPage": 1,
                    "pageSize": 10,
                    "gridConnectionTime": "",
                    "queryTime": round(time.time() * 1000),
                    "timeZone": 1,
                    "sortId": "createTime",
                    "sortDir": "DESC",
                    "locale": "en_US",
                }
                url = self.api_base + "/rest/pvms/web/station/v1/station/station-list"
                async with session.post(
                    url,
                    json=body,
                    headers=headers,
                ) as resp3:
                    self.logger.debug(f"response from {url}: {resp3}")
                    response = await resp3.json()
                    self._stationname = response["data"]["list"][0]["dn"]

            return True

        except Exception as ex:
            self.logger.exception(f"Failed to get data: {ex}")
            return False

    async def _get_rest_data(self, path, params={}):
        current_utc_time = int(datetime.utcnow().timestamp() * 1000)
        path = path.replace("%STATION%", self._stationname).replace(
            "%CURRENT_UTC_TIME%", str(current_utc_time)
        )
        if self._inverter_station:
            path = path.replace("%INVERTER_STATION%", self._inverter_station)

        url = self.api_base + path
        headers = {
            "accept": (
                "text/html,application/xhtml+xml,application/xml;q=0.9,"
                "image/avif,image/webp,image/apng,*/*;"
                "q=0.8,application/signed-exchange;v=b3;q=0.9,"
                "application/json, text/javascript, */*; q=0.01"
            ),
            "accept-language": (
                "en-AT,en;q=0.9,de-AT;q=0.8,de;q=0.7,en-GB;q=0.6,en-US;q=0.5"
            ),
            "roarand": self._csrftoken,
            "cache-control": "max-age=0",
            "content-type": "application/json",
            "sec-ch-ua": (
                '" Not A;Brand";v="99", "Chromium";v="100", "Google Chrome";v="100"'
            ),
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Chrome OS"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "none",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
        }
        resp = await self._get(url, headers, params=params)
        return resp

    async def send_idle(self):
        url = self.api_base + "/rest/plat/smapp/v1/idle"

        headers = {
            "accept": "application/json",
            "accept-language": (
                "en-AT,en;q=0.9,de-AT;q=0.8," "de;q=0.7,en-GB;q=0.6,en-US;q=0.5"
            ),
            "content-type": "application/json",
            "sec-ch-ua": (
                '".Not/A)Brand";v="99", "Google Chrome";v="103", "Chromium";v="103"'
            ),
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"ChromeOS"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "x-non-renewal-active": "true",
            "x-non-renewal-session": "false",
            "x-requested-with": "XMLHttpRequest",
        }
        await self._get(url, headers)

    async def _get(self, url, headers, max_retries=5, params={}):
        try:
            response = {}
            async with aiohttp.ClientSession(
                trust_env=True, headers=headers, cookie_jar=self._cookies
            ) as session:
                retry = 1
                while retry < max_retries:
                    retry += 1
                    async with session.get(url, params=params) as resp:
                        if resp.status != 200:
                            # wait a few seconds before login
                            self.logger.error(
                                f"Response from {url} was {resp.status}, try to relogin"
                            )
                            await asyncio.sleep(30)
                            await self.login()
                            continue
                        else:
                            if resp.headers["Content-Type"].startswith("text/html"):
                                await self.login()
                                continue
                            else:
                                response = await resp.json()
                                break

            if "success" in response and response["success"] is True:
                if "data" in response:
                    return response["data"]
                return response

            return response
        except Exception as ex:
            self.logger.exception(f"Failed to get data from {url}: {ex}")
            return {}

    async def update(self):
        await self.update_station_detail()
        await self.update_station_real_kpi()
        await self.update_energy_balance()
        await self.update_energy_flow()
        await self.update_device_signals()
        # await self.send_idle()

    async def update_energy_flow(self):
        self._inverter_output_power = "-"
        self._from_grid = 0
        self._to_grid = 0
        self._electrical_load = "-"
        self._string_output_power = "-"
        self._battery_soc = 0
        self._battery_battery_power = 0
        self._battery_charge_capacity = 0
        self._battery_discharge_capacity = 0
        self._energyflowdata = None

        # https://region01eu5.fusionsolar.huawei.com/rest/pvms/web/station/v1/overview/energy-flow?stationDn=STATION&_
        result = await self._get_rest_data(FusionSolarRestApi.ENERGY_FLOW_PATH)
        self._energyflowdata = result
        if "flow" in result:
            for node in result["flow"]["nodes"]:
                if node["name"] == "neteco.pvms.devTypeLangKey.inverter":
                    if node["deviceTips"]["ACTIVE_POWER"] == "--":
                        self._inverter_output_power = 0
                    else:
                        self._inverter_output_power = float(
                            node["deviceTips"]["ACTIVE_POWER"]
                        )
                    self._inverter_station = node["devIds"][0]
                elif node["name"] == "neteco.pvms.KPI.kpiView.electricalLoad":
                    self._electrical_load = float(
                        node["description"]["value"].replace(" kW", "")
                    )
                elif node["name"] == "neteco.pvms.devTypeLangKey.energy_store":
                    if node["deviceTips"]["SOC"] != "--":
                        self._battery_soc = float(node["deviceTips"]["SOC"])
                        self._battery_charge_capacity = float(
                            node["deviceTips"]["CHARGE_CAPACITY"]
                        )
                        self._battery_battery_power = float(
                            node["deviceTips"]["BATTERY_POWER"]
                        )
                        self._battery_discharge_capacity = float(
                            node["deviceTips"]["DISCHARGE_CAPACITY"]
                        )
                elif node["name"] == "neteco.pvms.devTypeLangKey.string":
                    self._string_output_power = float(
                        node["description"]["value"].replace(" kW", "")
                    )
            for node in result["flow"]["links"]:
                if "description" in node and "label" in node["description"]:
                    if (
                        node["description"]["label"]
                        == "neteco.pvms.energy.flow.buy.power"
                    ):
                        if node["description"]["value"] == "--":
                            self._to_grid = 0
                            self._from_grid = 0
                        else:
                            if node["fromNode"] == "3":
                                # from grid is fromNode 3
                                self._from_grid = float(
                                    node["description"]["value"].replace(" kW", "")
                                )
                                self._to_grid = 0
                            else:
                                # to grid is fromNode 2
                                self._to_grid = float(
                                    node["description"]["value"].replace(" kW", "")
                                )
                                self._from_grid = 0

    async def update_energy_balance(self):
        self._energy_balance = await self._get_rest_data(
            FusionSolarRestApi.ENERGY_BALANCE
        )

    async def update_available_device_signals(self):
        available_signals = await self._get_rest_data(
            FusionSolarRestApi.DEVICE_STATISTICS_SIGNALS
        )
        data = available_signals["signalList"]
        self._available_signals = SignalSet(
            signals=[
                Signal(row["id"], row["name"], row["unit"].get("unit", ""))
                for row in data
            ]
        )

    async def update_device_signals(self):
        if self._available_signals is None:
            await self.update_available_device_signals()

        self._device_signals = await self._get_rest_data(
            FusionSolarRestApi.DEVICE_SIGNALS,
            params={
                "deviceDn": self._inverter_station,
                "signalIds": self._available_signals.get_ids(),
                "date": int(datetime.now().timestamp() * 1000),
            },
        )

    async def update_station_detail(self):
        # https://region01eu5.fusionsolar.huawei.com/rest/pvms/web/station/v1/overview/station-detail?stationDn=STATION&_=
        self._stationdetail = await self._get_rest_data(
            FusionSolarRestApi.STATION_DETAIL_PATH
        )

    async def update_station_real_kpi(self):
        self._stationrealkpi = await self._get_rest_data(
            FusionSolarRestApi.STATION_REAL_KPI
        )

    @property
    def station(self):
        return self._stationdetail["dn"]

    @property
    def co2_saved(self):
        return self._stationdetail["co2"]

    @property
    def battery_soc(self):
        return self._battery_soc

    @property
    def battery_charge_capacity(self):
        return self._battery_charge_capacity

    @property
    def battery_power(self):
        return self._battery_battery_power

    @property
    def battery_discharge_capacity(self):
        return self._battery_discharge_capacity

    @property
    def total_lifetime_energy(self):
        return float(self._stationdetail["cumulativeEnergy"])

    @property
    def total_current_day_energy(self):
        return float(self._stationrealkpi["dailyEnergy"])

    @property
    def total_current_month_energy(self):
        return float(self._stationdetail["monthEnergy"])

    @property
    def total_current_year_energy(self):
        return float(self._stationdetail["yearEnergy"])

    @property
    def grid_connected_time(self):
        return self._stationdetail["gridConnectedTime"]

    @property
    def installed_capacity(self):
        return self._stationdetail["installedCapacity"]

    @property
    def daily_self_use_energy(self):
        return self._stationrealkpi["dailySelfUseEnergy"]

    @property
    def daily_use_energy(self):
        return self._stationrealkpi["dailyUseEnergy"]

    @property
    def daily_self_use_ratio(self):
        return self._energy_balance["selfUsePowerRatioByUse"]

    @property
    def daily_self_use_solar_ratio(self):
        return self._energy_balance["selfUsePowerRatioByProduct"]

    @property
    def inverter_output_power(self):
        return self._inverter_output_power

    @property
    def string_output_power(self):
        return self._string_output_power

    @property
    def electrical_load(self):
        return self._electrical_load

    @property
    def grid_power(self):
        if self._to_grid == 0 and self._from_grid == 0:
            return 0

        if self._to_grid > 0:
            return -(self._to_grid)

        return self._from_grid

    @property
    def to_grid_power(self):
        return self._to_grid

    @property
    def from_grid_power(self):
        return self._from_grid

    @property
    def device_signals(self):
        device_signals = {}
        for sigid in self._device_signals:
            sig_name = self._available_signals.get_name_by_id(sigid)
            sig_value_history = self._device_signals[sigid]["pmDataList"]
            last_val = "-"
            for row in sig_value_history:
                if "dnId" not in row:
                    continue
                last_val = row["counterValue"]
            device_signals[sig_name] = last_val
        return device_signals

    @property
    def data_string(self):
        return (
            f"Stationdetail Data:\n{self._stationdetail}\n\n"
            f"Energyflow Data:\n{self._energyflowdata}\n\n"
            f"Energybalance Data:\n{self._energy_balance}\n\n"
        )



import asyncio
import json
import logging
from typing import cast

import aiohomekit
from aiohomekit.model.characteristics.characteristic import NUMBER_TYPES
from aiohomekit.model.characteristics.characteristic_types import CharacteristicsTypes
from zeroconf import IPVersion, ServiceStateChange
from zeroconf.asyncio import (
    AsyncServiceBrowser,
    AsyncServiceInfo,
    AsyncZeroconfServiceTypes,
)

from .. import fhem
from .. import fhem_pythonbinding as fhempy
from .. import generic, utils
from ..core import zeroconf

HOMEKIT_SERVICES = ["_hap._udp.local.", "_hap._tcp.local."]


class homekit(generic.FhemModule):

    controller: aiohomekit.Controller = None
    aiobrowser: AsyncServiceBrowser = None

    def async_on_service_state_change(
        zeroconf,
        service_type: str,
        name: str,
        state_change: ServiceStateChange,
    ) -> None:
        if service_type in HOMEKIT_SERVICES:
            # print(
            #    f"Service {name} of type {service_type} state changed: {state_change}"
            # )
            if state_change is not ServiceStateChange.Added:
                return
            asyncio.ensure_future(
                homekit.async_display_service_info(zeroconf, service_type, name)
            )

    async def async_display_service_info(
        zeroconf, service_type: str, name: str
    ) -> None:
        info = AsyncServiceInfo(service_type, name)
        await info.async_request(zeroconf, 3000)
        # print("Info from zeroconf.get_service_info: %r" % (info))
        if info:
            addresses = [
                "%s:%d" % (addr, cast(int, info.port))
                for addr in info.parsed_scoped_addresses()
            ]
            # print("  Name: %s" % name)
            # print("  Addresses: %s" % ", ".join(addresses))
            # print("  Weight: %d, priority: %d" % (info.weight, info.priority))
            # print(f"  Server: {info.server}")
        # else:
        #    print("  No info")
        # print("\n")
        discovery_info = info
        properties = {
            key.lower().decode(): value
            for (key, value) in discovery_info.properties.items()
        }
        # print(f"props: {properties}")
        status_flags = int(properties["sf"])
        paired = not status_flags & 0x01
        # print(f"paired: {paired}")
        hkid = properties["id"].lower()
        model = properties["md"]

    async def get_controller():
        if homekit.controller is None:
            aio_zc = zeroconf.zeroconf.get_instance(
                logging.Logger("homekit")
            ).get_async_zeroconf()
            services = HOMEKIT_SERVICES
            homekit.aiobrowser = AsyncServiceBrowser(
                aio_zc.zeroconf,
                services,
                handlers=[homekit.async_on_service_state_change],
            )
            homekit.controller = aiohomekit.Controller(aio_zc)
            await homekit.controller.async_start()
        return homekit.controller

    def __init__(self, logger):
        self._ready = asyncio.Event()
        super().__init__(logger)

    @property
    def ready(self):
        return self._ready

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "interval": {
                "default": 60,
                "format": "int",
                "help": "Change interval, default is 60 seconds.",
            },
            "pairing_data": {
                "default": "{}",
                "format": "json",
                "help": "Pairing data from HomeKit device",
            },
        }
        await self.set_attr_config(attr_config)

        self.set_config = {}
        await self.set_set_config(self.set_config)

        if len(args) != 5:
            return (
                "Usage: define velux_gateway fhempy homekit HOMEKITID|GW_NAME PIN|AID"
            )
        if args[4].find("-") > -1:
            # homekit pin found
            hash["HOMEKIT_ID"] = args[3]
            hash["HOMEKIT_PIN"] = args[4]
            self.create_async_task(
                self.setup_gateway_device(hash["HOMEKIT_ID"], args[4])
            )
        else:
            hash["HOMEKIT_GW"] = args[3]
            hash["HOMEKIT_AID"] = args[4]
            self.create_async_task(self.setup_device(args[3], int(args[4])))

    async def set_attr_pairing_data(self, hash):
        await self.setup_gateway_device(hash["HOMEKIT_ID"], hash["HOMEKIT_PIN"])

    async def setup_gateway_device(self, homekitid, pin):
        controller = await homekit.get_controller()

        self.pairing = None
        if self._attr_pairing_data == {}:
            # not paired yet
            discovery = await controller.async_find(homekitid)
            finish_pairing = await discovery.async_start_pairing(homekitid)
            self.pairing = await finish_pairing(pin)
            pairing_data = json.dumps(self.pairing.pairing_data)
            await fhem.CommandAttr(
                self.hash, f"{self.hash['NAME']} pairing_data {pairing_data}"
            )
        else:
            self.pairing = controller.load_pairing(
                self._attr_pairing_data["AccessoryPairingID"],
                dict(self._attr_pairing_data),
            )

        name = await self.pairing.get_primary_name()
        await fhem.readingsSingleUpdate(self.hash, "name", name, 1)

        accessories = await self.pairing.list_accessories_and_characteristics()
        for accessory in accessories:
            devname = utils.gen_fhemdev_name(
                f"hkdev_{self.hash['HOMEKIT_ID']}_{accessory['aid']}"
            )
            dev_ex = await fhem.checkIfDeviceExists(
                self.hash,
                "HOMEKIT_GW",
                self.hash["NAME"],
                "HOMEKIT_AID",
                str(accessory["aid"]),
            )
            if dev_ex:
                continue

            await fhem.CommandDefine(
                self.hash,
                f"{devname} fhempy homekit {self.hash['NAME']} {accessory['aid']}",
            )

        self._ready.set()

    async def setup_device(self, gw_device_name, aid):
        controller = await homekit.get_controller()
        # load pairing data from homekit gateway
        gw_device = fhempy.getFhemPyDeviceByName(gw_device_name)
        await gw_device.ready.wait()
        self.pairing = controller.load_pairing(
            gw_device._attr_pairing_data["AccessoryPairingID"],
            dict(gw_device._attr_pairing_data),
        )
        gw_name = await self.pairing.get_primary_name()
        await fhem.readingsSingleUpdateIfChanged(self.hash, "gw_name", gw_name, 1)

        for accessory in self.pairing.accessories:
            if accessory.aid == aid:
                for service in accessory.services:
                    for char in service.characteristics:
                        if "pw" in char.perms:
                            # char.type convert to string (position_target)
                            await self.char_to_set(char)
                        if "pr" in char.perms:
                            if char.description:
                                reading = char.description
                            else:
                                reading = char.type
                            await fhem.readingsSingleUpdateIfChanged(
                                self.hash,
                                utils.gen_reading_name(reading),
                                char.value,
                                1,
                            )

        await self.set_set_config(self.set_config)

    async def set_int(self, hash, params):
        char = params["function_param"]
        self.create_async_task(
            self.pairing.put_characteristics(
                [(int(self.hash["HOMEKIT_AID"]), char.iid, params["new_value"])]
            )
        )

    async def char_to_set(self, char):
        if char.format in NUMBER_TYPES:
            self.set_config[utils.gen_reading_name(char.description)] = {
                "args": ["new_value"],
                "params": {"new_value": {"format": "int"}},
                "function": "set_int",
                "function_param": char,
                "options": f"slider,{char.minValue},{char.minStep},{char.maxValue},1",
            }



import asyncio
import functools
import json
from io import BytesIO

from tuya_sharing import (
    CustomerDevice,
    LoginControl,
    Manager,
    SharingDeviceListener,
    SharingTokenListener,
)

import fhempy.lib.fhem as fhem
import fhempy.lib.utils as utils

from .const import APP_QR_CODE_HEADER, CONF_CLIENT_ID, CONF_SCHEMA


class tuya_smartlife_setup:
    def __init__(self, logger, fhemdevice):
        self.logger = logger
        self.fhemdev = fhemdevice
        self.usercode = self.fhemdev.usercode
        self.hash = fhemdevice.hash
        self._t_devicelist = []
        self._ready = False

        # init login control
        self.login_control = LoginControl()

    async def login(self):
        # use LoginControl.qr_code to login

        # get token_info from reading
        self.token_info = json.loads(
            await fhem.ReadingsVal(self.hash["NAME"], "token_info", "{}")
        )
        # get terminal_id from reading
        self.terminal_id = await fhem.ReadingsVal(self.hash["NAME"], "terminal_id", "")
        # get endpoint from reading
        self.endpoint = await fhem.ReadingsVal(self.hash["NAME"], "endpoint", "")

        # check if token_info is not {}
        if self.token_info == {}:
            response = await utils.run_blocking(
                functools.partial(
                    self.login_control.qr_code,
                    CONF_CLIENT_ID,
                    CONF_SCHEMA,
                    self.usercode,
                )
            )

            # check if response has success True
            if response.get("success", False):
                qr_code = response["result"]["qrcode"]
                self._qr_code = qr_code
                self.logger.debug("qr_code=%s", qr_code)
                img = self._generate_qr_code(APP_QR_CODE_HEADER + qr_code)

                # set reading qr_code to img
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "login_qr_code", "<html>" + img + "</html>", 1
                )

                # set attr webCmd to scan_done
                await fhem.CommandAttr(
                    self.hash, f"{self.hash['NAME']} webCmd scan_done"
                )

                # inform user via state to scan qr_code and press scan_done
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash,
                    "state",
                    "Please scan QR Code and press scan_done",
                    1,
                )
        else:
            # connect to smartlife
            self.fhemdev.create_async_task(self.connect_to_smartlife())

    async def set_scan_done(self, hash, params):
        """Check login_result from LoginControl to get token and secret"""
        ret, info = await utils.run_blocking(
            functools.partial(
                self.login_control.login_result,
                self._qr_code,
                CONF_CLIENT_ID,
                self.usercode,
            )
        )
        if ret:
            self.token_info = {
                "t": info.get("t"),
                "uid": info.get("uid"),
                "expire_time": info.get("expire_time"),
                "access_token": info.get("access_token"),
                "refresh_token": info.get("refresh_token"),
            }

            self.terminal_id = info.get("terminal_id")
            self.endpoint = info.get("endpoint")

            # remove login_qr_code reading and webCmd attr
            await fhem.readingsSingleUpdateIfChanged(self.hash, "login_qr_code", "-", 1)
            await fhem.CommandDeleteAttr(self.hash, f"{self.hash['NAME']} webCmd")

            await self.save_token_info()

            self.fhemdev.create_async_task(self.connect_to_smartlife())
        else:
            self.logger.error(f"login_result failed: {info}")
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "state", f"login failed: {info}", 1
            )

    async def save_token_info(self):
        # save token_info to reading
        await fhem.readingsSingleUpdateIfChanged(
            self.hash,
            "token_info",
            json.dumps(self.token_info),
            1,
        )

        # save terminal_id and endpoint to reading
        await fhem.readingsSingleUpdateIfChanged(
            self.hash,
            "terminal_id",
            self.terminal_id,
            1,
        )
        await fhem.readingsSingleUpdateIfChanged(
            self.hash,
            "endpoint",
            self.endpoint,
            1,
        )

    async def connect_to_smartlife(self):
        token_listener = TokenListener(self.logger, self)
        self.device_manager = Manager(
            CONF_CLIENT_ID,
            self.usercode,
            self.terminal_id,
            self.endpoint,
            self.token_info,
            token_listener,
        )

        t_cloud_setup = self

        class DeviceListener(SharingDeviceListener):
            """Device Update Listener."""

            def __init__(self, logger, manager) -> None:
                super().__init__()
                self.logger = logger
                self.manager = manager

            def update_device(self, device: CustomerDevice):
                self.logger.debug(f"update_device received for {device.id}")
                for dev in t_cloud_setup.tuya_devices:
                    if dev.id == device.id:
                        try:
                            asyncio.run_coroutine_threadsafe(
                                dev.update(device), t_cloud_setup.fhemdev.loop
                            )
                        except Exception:
                            self.logger.exception("Failed to update device")

            def add_device(self, device: CustomerDevice):
                self.logger.info(f"add_device received for {device.id}")
                try:
                    asyncio.run_coroutine_threadsafe(
                        self.add_fhem_device(device), t_cloud_setup.fhemdev.loop
                    )
                except Exception:
                    self.logger.exception("Failed to add device")

            async def add_fhem_device(self, device: CustomerDevice):
                await t_cloud_setup._create_fhem_device(device.name, device.id)

            def remove_device(self, device_id: str):
                self.logger.info(f"remove_device received for {device_id}")

        await utils.run_blocking(
            functools.partial(self.device_manager.update_device_cache)
        )

        listener = DeviceListener(self.logger, self.device_manager)
        self.device_manager.add_device_listener(listener)

        self._ready = True

        await self._init_devices()

        await utils.run_blocking(functools.partial(self.device_manager.refresh_mq))
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connected", 1)

    def _generate_qr_code(self, data: str) -> str:
        """Generate a base64 PNG string represent QR Code image of data."""
        import pyqrcode  # pylint: disable=import-outside-toplevel

        qr_code = pyqrcode.create(data)

        with BytesIO() as buffer:
            qr_code.svg(file=buffer, scale=4)
            return str(
                buffer.getvalue()
                .decode("ascii")
                .replace("\n", "")
                .replace(
                    (
                        '<?xml version="1.0" encoding="UTF-8"?>'
                        '<svg xmlns="http://www.w3.org/2000/svg"'
                    ),
                    "<svg style='width:228px;height:228px;'",
                )
            )

    async def Undefine(self, hash):
        return

    @property
    def ready(self):
        return self._ready

    def register_tuya_device(self, device):
        self._t_devicelist.append(device)

    def unregister_tuya_device(self, device):
        self._t_devicelist.remove(device)

    @property
    def tuya_devices(self):
        return self._t_devicelist

    async def _create_fhem_device(self, name, device_id):
        devalias = name
        devname = "tuya_smartlife_" + device_id
        devname = utils.gen_fhemdev_name(devname)
        device_exists = await fhem.checkIfDeviceExists(
            self.hash, "FHEMPYTYPE", "tuya_smartlife", "DEVICEID", device_id
        )

        if not device_exists:
            self.logger.info(
                (
                    f"create: {devname} fhempy tuya_smartlife "
                    f"{self.hash['NAME']} {device_id}"
                )
            )
            # define each device (CommandDefine ... tuya_cloud_setup_dev deviceid
            await fhem.CommandDefine(
                self.hash,
                (
                    f"{devname} fhempy tuya_smartlife "
                    f"{self.hash['NAME']} {device_id}"
                ),
            )
            await fhem.CommandAttr(self.hash, f"{devname} alias {devalias}")
            # wait for FHEM to handle CommandDefine
            await asyncio.sleep(1)

    async def _init_devices(self):
        # wait for init to complete, otherwise devices might not be available yet
        while await fhem.init_done(self.hash) == 0:
            await asyncio.sleep(3)

        # retrieve devices from device_manager and create them
        for device_id in self.device_manager.device_map:
            await self._create_fhem_device(
                self.device_manager.device_map[device_id].name, device_id
            )
            await asyncio.sleep(1)

    async def send_commands(self, deviceid, commands):
        await utils.run_blocking(
            functools.partial(self.device_manager.send_commands, deviceid, commands)
        )


class TokenListener(SharingTokenListener):
    def __init__(self, logger, tuya_smartlife_setup: tuya_smartlife_setup) -> None:
        """Init TokenListener."""
        self.logger = logger
        self.tuya_smartlife_setup = tuya_smartlife_setup

    def update_token(self, token_info):
        self.logger.debug("update token info : %s", token_info)
        self.tuya_smartlife_setup.token_info = token_info
        asyncio.run_coroutine_threadsafe(
            self.tuya_smartlife_setup.save_token_info(),
            self.tuya_smartlife_setup.fhemdev.loop,
        )



"""
A simple wrapper for bleak
Handles reconnection
"""

import asyncio
import functools
import getpass
import os
import time

import aiofiles
import bluetooth_adapters
from bleak import BleakClient, BleakScanner
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError

from .. import fhem, utils
from .bluetoothctl import Bluetoothctl, PairingState


class BluetoothLE:
    # run bluetoothctl only once
    bluetoothctl_lock = None

    # keep connected (get manuf uuid every x seconds)
    # reset hci devices on errors
    # find by name
    # add bluetooth-auto-recovery
    def __init__(
        self,
        logger,
        device_hash,
        address=None,
        name=None,
        pairing_required=False,
        pin="0000",
    ) -> None:
        self._disconnect_called = False
        self._device = None
        self._client = None
        self._dev_hash = device_hash

        self.pairing_required = pairing_required
        self.pin = pin
        self.disconnect_listener = None
        self.connected_listener = None

        self.logger = logger
        self.addr = address
        self.name = name

        self.conf_checked = False
        self.paired = False

        self.connection_task = None
        self.connected = asyncio.Event()

        # initialize bluetoothctl_lock here to avoid thread without event loop error
        if BluetoothLE.bluetoothctl_lock is None:
            BluetoothLE.bluetoothctl_lock = asyncio.Lock()

    async def pair(self):
        if self.addr is None:
            return False

        await fhem.readingsSingleUpdateIfChanged(
            self._dev_hash, "connection_paired", "pairing", 1
        )

        async with BluetoothLE.bluetoothctl_lock:
            ret = await utils.run_blocking(functools.partial(self._pair, self.pin))

        # get enum text from PairingState from ret
        if ret == PairingState.SUCCESS:
            await fhem.readingsSingleUpdateIfChanged(
                self._dev_hash, "connection_paired", "paired", 1
            )
        elif ret == PairingState.FAILED:
            await fhem.readingsSingleUpdateIfChanged(
                self._dev_hash, "connection_paired", "failed", 1
            )
        elif ret == PairingState.WRONG_PIN:
            await fhem.readingsSingleUpdateIfChanged(
                self._dev_hash, "connection_paired", "wrong PIN", 1
            )
        elif ret == PairingState.TIMEOUT:
            await fhem.readingsSingleUpdateIfChanged(
                self._dev_hash, "connection_paired", "timeout", 1
            )

        return ret

    def _pair(self, pin, retry=3):
        btctl = Bluetoothctl(self.logger)

        paired_devices = btctl.get_paired_devices()
        if self.addr in [d["mac_address"] for d in paired_devices]:
            btctl.exit()
            return PairingState.SUCCESS

        while retry > 0:
            try:
                btctl.power_on()
                btctl.agent_on()
                btctl.default_agent()
                btctl.start_scan()
                time.sleep(10)
                btctl.stop_scan()
                pairing_successfull = btctl.pair(self.addr, pin)
                if pairing_successfull == PairingState.SUCCESS:
                    btctl.trust(self.addr)
                    btctl.disconnect(self.addr)
                    btctl.exit()
                    return pairing_successfull
                else:
                    retry -= 1
                    time.sleep(5)
            except Exception as e:
                self.logger.error(e)
                retry -= 1
                time.sleep(5)

        btctl.exit()

        return PairingState.FAILED

    async def update_adapters(self):
        self.adapters = await bluetooth_adapters.get_bluetooth_adapters()
        adapters = bluetooth_adapters.get_adapters()
        await adapters.refresh()
        self.adapter_details = adapters.adapters

    async def find_device(self, timeout=30, adapter=None):
        self._device = None

        async with BluetoothLE.bluetoothctl_lock:
            try:
                self._device = await BleakScanner.find_device_by_address(
                    self.addr, timeout=timeout, adapter=adapter
                )
                self.logger.info(f"Device found via adapter {adapter}")
            except (asyncio.TimeoutError, BleakError):
                # nothing found
                self._device = None

    def register_disconnect_listener(self, disconnect_listener):
        self.disconnect_listener = disconnect_listener

    def register_connection_established_listener(self, connected_listener):
        self.connected_listener = connected_listener

    def register_notification_listener(self, notification_listener):
        self.notification_listener = notification_listener

    async def connect(self, timeout=30, max_retries=20):
        """Connect to the device."""

        # check which user is running this code
        user = getpass.getuser()

        # check if bluetooth.conf is present and contains the required policy lines
        if not self.conf_checked:
            btconf = "/etc/dbus-1/system.d/bluetooth.conf"
            policy_lines = [
                '<policy user="' + user + '">',
                '<allow own="org.bluez"/>',
                '<allow send_destination="org.bluez"/>',
                '<allow send_interface="org.bluez.GattCharacteristic1"/>',
                '<allow send_interface="org.bluez.GattDescriptor1"/>',
                '<allow send_interface="org.freedesktop.DBus.ObjectManager"/>',
                '<allow send_interface="org.freedesktop.DBus.Properties"/>',
                "</policy>",
            ]
            if os.path.exists(btconf):
                async with aiofiles.open(btconf, mode="r") as f:
                    content = await f.read()
                    if not all(line in content for line in policy_lines):
                        self.logger.error(
                            "Not all required policy lines are present in bluetooth.conf"
                        )
                        self.logger.error(
                            "Please add the following lines to the file /etc/dbus-1/system.d/bluetooth.conf:"
                        )
                        for line in policy_lines:
                            self.logger.error(line)
                        return
            self.conf_checked = True

        if self.pairing_required and not self.paired:
            await fhem.readingsSingleUpdateIfChanged(
                self._dev_hash, "connection", "pairing", 1
            )
            ret = await self.pair()
            if ret:
                self.paired = True

        if self._client and self._client.is_connected:
            return

        if self.connection_task is None or self.connection_task.done():
            self.connection_task = asyncio.create_task(
                self.connect_loop(timeout, max_retries)
            )

    async def connect_loop(self, timeout, max_retries):
        while True:
            try:
                await self.connect_once(timeout, max_retries)
                if self._client and self._client.is_connected:
                    # loop is started again on disconnect
                    self.connected.set()
                    return
            except asyncio.CancelledError:
                return
            except Exception:
                self.logger.exception("Failed to connect")

            # wait 10s before next retry
            await asyncio.sleep(10)

    async def connect_once(self, timeout, max_retries):
        # get latest adapter list
        await self.update_adapters()

        await fhem.readingsSingleUpdateIfChanged(
            self._dev_hash, "connection", "connecting", 1
        )
        self._disconnect_called = False

        for i in range(0, max_retries):
            for adapter in self.adapters:
                await self.find_device(timeout=timeout, adapter=adapter)
                if not self._device:
                    # device not found
                    await asyncio.sleep(20)
                    continue

                await fhem.readingsSingleUpdateIfChanged(
                    self._dev_hash, "connection_adapter", adapter, 1
                )
                manufacturer = self.adapter_details[adapter]["manufacturer"]
                if not manufacturer:
                    manufacturer = "unknown"
                await fhem.readingsSingleUpdateIfChanged(
                    self._dev_hash,
                    "connection_adapter_details",
                    manufacturer
                    + " ("
                    + self.adapter_details[adapter]["address"]
                    + ")",
                    1,
                )
                await fhem.readingsSingleUpdateIfChanged(
                    self._dev_hash, "rssi", self._device.rssi, 1
                )

                self._client = BleakClient(
                    self._device,
                    disconnected_callback=self._disconnect_callback,
                    timeout=timeout,
                    adapter=adapter,
                )
                try:
                    await self._client.connect()
                except BleakError:
                    self.logger.exception("Failed to connect")
                except asyncio.TimeoutError:
                    pass

                if self._client and self._client.is_connected:
                    try:
                        await self._subscribe_notifies()
                    except BleakError:
                        # set state to connected (no notifications)
                        await fhem.readingsSingleUpdate(
                            self._dev_hash,
                            "connection",
                            "connected (no notifications)",
                            1,
                        )
                        self.logger.exception("Failed to subscribe to notifications")
                    else:
                        await fhem.readingsSingleUpdate(
                            self._dev_hash, "connection", "connected", 1
                        )
                    if self.connected_listener:
                        await self.connected_listener()
                    return
                else:
                    await asyncio.sleep(20)
                    continue

        # connection failed
        return

    async def _subscribe_notifies(self):
        if not self.notification_listener:
            return

        services = self._client.services
        for service in services:
            for characteristic in service.characteristics:
                if "notify" in characteristic.properties:
                    self.logger.debug(f"start_notify for {characteristic.uuid}")
                    await self._client.start_notify(
                        characteristic.uuid, self.notification_listener
                    )

    async def write_gatt_char(self, uuid, data):
        await asyncio.wait_for(self.connected.wait(), 30)
        if self.connected.is_set():
            await self._client.write_gatt_char(uuid, data)

    async def read_gatt_char(self, uuid):
        await asyncio.wait_for(self.connected.wait(), 30)
        if self.connected.is_set():
            return await self._client.read_gatt_char(uuid)

    async def disconnect(self):
        self._disconnect_called = True

        if self.connection_task:
            self.connection_task.cancel()

        if self._client and self._client.is_connected:
            await self._client.disconnect()

    def _disconnect_callback(self, client: BleakClient):
        self.connected.clear()
        asyncio.create_task(self.async_disconnected(client))

    async def async_disconnected(self, client: BleakClient):
        await fhem.readingsSingleUpdate(self._dev_hash, "connection", "disconnected", 1)
        if self.disconnect_listener:
            await self.disconnect_listener(client)

        self._client = None
        self._device = None

        if not self._disconnect_called:
            self.logger.error("Device disconnected, reconnect now")
            await self.connect()
        else:
            self.logger.info("Device disconnected")

    @property
    def is_connected(self) -> bool:
        return self._client is not None and self._client.is_connected

    @property
    def device(self) -> BLEDevice:
        return self._device

    @property
    def client(self) -> BleakClient:
        return self._client



#!/usr/bin/env python3
"""Support for Tuya Smart devices."""

import logging

import fhempy.lib.tuya_cloud.tuya_cloud_device as tcd
import fhempy.lib.tuya_cloud.tuya_cloud_setup as tcs
from fhempy.lib import fhem
from fhempy.lib.generic import FhemModule


class tuya_cloud(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.device = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 5 or (len(args) < 8 and args[3] == "setup"):
            return (
                "Usage: define tuya_cloud fhempy tuya_cloud"
                " setup <API_KEY> <API_SECRET> <USERNAME> <PASSWORD>"
                " [<APPTYPE=smartlife>] [<REGION>=Europe]<br>"
                "OR if you want to define only one device<br>"
                " <TUYA_SETUP_DEVICE> <DEVICEID>"
            )

        if args[3] == "setup":
            self.device = tcs.tuya_cloud_setup(self.logger, self)
            if await fhem.AttrVal(self.hash["NAME"], "verbose", "3") == "5":
                logging.getLogger("tuya_iot").setLevel(logging.DEBUG)
            else:
                logging.getLogger("tuya_iot").setLevel(logging.ERROR)
        else:
            self.device = tcd.tuya_cloud_device(self.logger, self)

        await self.device.Define(self.hash, args, argsh)

    async def Undefine(self, hash):
        if self.device:
            await self.device.Undefine(hash)
        return await super().Undefine(hash)

    async def set_reset_energy(self, hash, params):
        await self.device.set_reset_energy(hash, params)

    async def set_boolean(self, hash, params):
        await self.device.set_boolean(hash, params)

    async def set_enum(self, hash, params):
        await self.device.set_enum(hash, params)

    async def set_json(self, hash, params):
        await self.device.set_json(hash, params)

    async def set_string(self, hash, params):
        await self.device.set_string(hash, params)

    async def set_integer(self, hash, params):
        await self.device.set_integer(hash, params)

    async def set_colour_data(self, hash, params):
        await self.device.set_colour_data(hash, params)

    async def set_colour_data_v2(self, hash, params):
        await self.device.set_colour_data_v2(hash, params)

    @property
    def tuya_cloud_device(self):
        return self.device



import asyncio

import aiohttp

from .. import fhem, generic

import datetime as dt


class volvo(generic.FhemModule):

    VEHICLELIST = "https://api.volvocars.com/extended-vehicle/v1/vehicles"
    VEHICLE_COMMANDS = (
        "https://api.volvocars.com/connected-vehicle/v1/vehicles/{vin}/commands"
    )
    VEHICLE_COMMANDS_ACCEPT = (
        "application/vnd.volvocars.api.connected-vehicle.commandlist.v1+json"
    )

    REGULAR_UPDATE_URLS = {
        "https://api.volvocars.com/energy/v1/vehicles/{vin}/recharge-status": "application/vnd.volvocars.api.energy.vehicledata.v1+json",
        "https://api.volvocars.com/connected-vehicle/v1/vehicles/{vin}/doors": "application/vnd.volvocars.api.connected-vehicle.vehicledata.v1+json",
        "https://api.volvocars.com/connected-vehicle/v1/vehicles/{vin}/environment": "application/vnd.volvocars.api.connected-vehicle.vehicledata.v1+json",
        "https://api.volvocars.com/connected-vehicle/v1/vehicles/{vin}/engine": "application/vnd.volvocars.api.connected-vehicle.vehicledata.v1+json",
        "https://api.volvocars.com/connected-vehicle/v1/vehicles/{vin}/engine-status": "application/vnd.volvocars.api.connected-vehicle.vehicledata.v1+json",
        "https://api.volvocars.com/connected-vehicle/v1/vehicles/{vin}/odometer": "application/vnd.volvocars.api.connected-vehicle.vehicledata.v1+json",
        "https://api.volvocars.com/connected-vehicle/v1/vehicles/{vin}/statistics": "application/vnd.volvocars.api.connected-vehicle.vehicledata.v1+json",
        "https://api.volvocars.com/connected-vehicle/v1/vehicles/{vin}/tyres": "application/vnd.volvocars.api.connected-vehicle.vehicledata.v1+json",
        "https://api.volvocars.com/connected-vehicle/v1/vehicles/{vin}/warnings": "application/vnd.volvocars.api.connected-vehicle.vehicledata.v1+json",
    }

    def __init__(self, logger):
        super().__init__(logger)

        self.session = None
        self.access_token = None
        self.refresh_token = None
        self.vin = "NO_VIN"

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        

        self.session = aiohttp.ClientSession()

        self.attr_config = {
            "car": {
                "default": "",
                "help": "Select the car you would like to control.",
            },
            "interval": {
                "default": 300,
                "format": "int",
                "help": "Readings update intervall in seconds (default 5min).",
            },
        }
        await self.set_attr_config(self.attr_config)
        await self.set_icon("car")

        if len(args) != 6:
            return "Usage: define my_volvo fhempy volvo appkey username password"
        self.appkey = args[3]
        self.username = args[4]
        self.password = args[5]
        self.create_async_task(self.update_loop())

    async def Undefine(self, hash):
        if self.session:
            await self.session.close()

    async def login(self):
        url = "https://volvoid.eu.volvocars.com/as/token.oauth2"
        headers = {
            "content-type": "application/x-www-form-urlencoded",
            "authorization": (
                "Basic "
                "aDRZZjBiOlU4WWtTYlZsNnh3c2c1WVFxWmZyZ1ZtSWFEcG"
                "hPc3kxUENhVXNpY1F0bzNUUjVrd2FKc2U0QVpkZ2ZJZmNMeXc="
            ),
            "access_token_manager_id": "JWTh4Yf0b",
            "user-agent": "okhttp/4.10.0",
        }
        params = {
            "username": self.username,
            "password": self.password,
            "grant_type": "password",
            "scope": " ".join(
                [
                    "openid",
                    "email",
                    "profile",
                    "care_by_volvo:financial_information:invoice:read",
                    "care_by_volvo:financial_information:payment_method",
                    "care_by_volvo:subscription:read",
                    "customer:attributes",
                    "customer:attributes:write",
                    "order:attributes",
                    "vehicle:attributes",
                    "tsp_customer_api:all",
                    "conve:brake_status",
                    "conve:climatization_start_stop",
                    "conve:command_accessibility",
                    "conve:commands",
                    "conve:diagnostics_engine_status",
                    "conve:diagnostics_workshop",
                    "conve:doors_status",
                    "conve:engine_status",
                    "conve:environment",
                    "conve:fuel_status",
                    "conve:honk_flash",
                    "conve:lock",
                    "conve:lock_status",
                    "conve:navigation",
                    "conve:odometer_status",
                    "conve:trip_statistics",
                    "conve:tyre_status",
                    "conve:unlock",
                    "conve:vehicle_relation",
                    "conve:warnings",
                    "conve:windows_status",
                    "energy:battery_charge_level",
                    "energy:charging_connection_status",
                    "energy:charging_system_status",
                    "energy:electric_range",
                    "energy:estimated_charging_time",
                    "energy:recharge_status",
                    "vehicle:attributes",
                ]
            ),
        }
        try:
            response = {}
            async with self.session.post(url, headers=headers, data=params) as resp:
                if resp.content_type != "application/json":
                    await fhem.readingsSingleUpdate(self.hash, "state", "Login failed")
                    return

                response = await resp.json()

                if resp.status == 200:
                    self.access_token = response["access_token"]
                    self.refresh_token = response["refresh_token"]
                    self.expires_in = response["expires_in"]
                else:
                    self.logger.error(f"Failed to get data from {url}: {response}")
        except Exception:
            self.logger.exception(f"Failed to get data from {url}")

    async def update_with_refresh_token(self):
        url = "https://volvoid.eu.volvocars.com/as/token.oauth2"
        headers = {
            "content-type": "application/x-www-form-urlencoded",
            "authorization": (
                "Basic "
                "aDRZZjBiOlU4WWtTYlZsNnh3c2c1WVFxWmZyZ1ZtSWFEcG"
                "hPc3kxUENhVXNpY1F0bzNUUjVrd2FKc2U0QVpkZ2ZJZmNMeXc="
            ),
            "user-agent": "okhttp/4.10.0",
        }
        params = {
            "access_token_manager_id": "JWTh4Yf0b",
            "grant_type": "refresh_token",
            "refresh_token": self.refresh_token,
        }
        try:
            response = {}
            async with self.session.post(url, headers=headers, data=params) as resp:
                if resp.content_type != "application/json":
                    await fhem.readingsSingleUpdate(self.hash, "state", "Login failed")
                    return

                response = await resp.json()

                if resp.status == 200:
                    self.access_token = response["access_token"]
                    self.refresh_token = response["refresh_token"]
                    self.expires_in = response["expires_in"]
                else:
                    self.logger.error(f"Failed to get data from {url}: {response}")
        except Exception:
            self.logger.exception(f"Failed to get data from {url}")

    async def update_loop(self):
        await fhem.readingsSingleUpdate(self.hash, "state", "connecting", 1)
        await self.login()
        if self.access_token is None:
            await fhem.readingsSingleUpdate(self.hash, "state", "connection failed", 1)
            return

        self.create_async_task(self.update_token())
        await fhem.readingsSingleUpdate(self.hash, "state", "connected", 1)

        await self.get_cars()
        await self.get_commands()

        while True:
            await self.get_regular_update_urls()
            await asyncio.sleep(self._attr_interval)

    async def get_commands(self):
        try:
            cmds = await self.volvo_get(
            volvo.VEHICLE_COMMANDS, volvo.VEHICLE_COMMANDS_ACCEPT
            )
        except Exception:
            self.logger.exception("Failed to get commands")
            return
        
        set_conf = {}
        for cmd in cmds["data"]:
            set_conf[cmd["command"].lower()] = {
                "function": "set_command",
                "function_param": {"command": cmd["command"], "href": cmd["href"]},
            }
        await self.set_set_config(set_conf)

    async def set_command(self, hash, params):
        cmd = params["function_param"]["command"]
        url = params["function_param"]["href"]
        self.create_async_task(self.execute_command(cmd, url))

    async def execute_command(self, command, url):
        cmd_type = command.replace("_", "").lower()
        headers = {
            "content-type": f"application/vnd.volvocars.api.connected-vehicle.{cmd_type}.v1+json",
            "vcc-api-key": self.appkey,
            "authorization": f"Bearer {self.access_token}",
        }
        try:
            response = {}
            data = None
            if cmd_type == "unlock":
                data = {"unlockDuration": 120}
            async with self.session.post(url, headers=headers, data=data) as resp:
                response = await resp.json()

                if response["status"] == 202:
                    await asyncio.sleep(10)
                    async_href = response["async"]["href"]
                    resp = await self.volvo_get(
                        async_href,
                        "application/vnd.volvocars.api.connected-vehicle.requestdetailresponse.v1+json",
                    )
                    if resp["status"] == 200:
                        await fhem.readingsSingleUpdate(
                            self.hash, "lastcommand", "ok", 1
                        )
                    else:
                        await fhem.readingsSingleUpdate(
                            self.hash, "lastcommand", "failed", 1
                        )
        except Exception:
            self.logger.exception(f"Failed to get data from {url}")
            return {}

    async def update_token(self):
        while True:
            await asyncio.sleep(self.expires_in - 30)
            await self.update_with_refresh_token()

    async def get_regular_update_urls(self):
        for url in volvo.REGULAR_UPDATE_URLS:
            start = url.rfind("/") + 1
            domain = url[start:].replace("-", "_")
            try:
                data = await self.volvo_get(url, volvo.REGULAR_UPDATE_URLS[url])
                await self.update_readings(data["data"], domain)
            except Exception:
                await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "failed to read data from URL", 1)
        
    async def get_cars(self):
        
        try:
            cars = await self.volvo_get(volvo.VEHICLELIST)
            carList = cars["vehicles"]
            if len(carList) == 0:
                self.logger.error("No cars found")
                return
            elif len(carList) == 1:
                self.vin = carList["id"]
            else:
                carVID = await fhem.AttrVal(self.hash["NAME"], "car", "")
                if (carVID == ""):
                    car_ids = [car["id"] for car in carList]
                    car_ids_str = ",".join(car_ids)
                    self.logger.info(f"Car IDs: {car_ids_str}")
                    await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "found cars: " + car_ids_str + " Please set the attribute car", 1)
                else:
                    self.vin = carVID
        except Exception:
            self.logger.exception("Failed to get cars")
            return
        

    async def update_readings(self, data, domain=""):
        try:
            for reading in data:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, domain + "_" + reading, data[reading]["value"], 1
                )
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash,
                    domain + "_" + reading + "_timestamp",
                    data[reading]["timestamp"],
                    1,
                )
        except Exception:
            self.logger.exception("Failed to update readings")

    async def volvo_get(self, url, accept_header="application/json"):
        url = url.replace("{vin}", self.vin)
        headers = {
            "accept": accept_header,
            "vcc-api-key": self.appkey,
            "authorization": f"Bearer {self.access_token}",
        }
        try:
            response = {}
            async with self.session.get(url, headers=headers) as resp:
                response = await resp.json()

                if resp.status == 200:
                    return response
                elif resp.status == 403:
                  await fhem.readingsSingleUpdateIfChanged(
                    self.hash,
                    "state",
                    response,
                    1,
                )  
                else:
                    self.logger.error(f"Failed to get data from {url}: {response}")
        except Exception:
            self.logger.exception(f"Failed to get data from {url}")
            return {}



from .. import fhem, generic
from ..core import bluetoothle


class mitemp2(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 4:
            return "Usage: define mitemp fhempy mitemp2 <MAC>"
        self._mac = args[3]
        self.hash["MAC"] = self._mac
        self._conn = bluetoothle.BluetoothLE(
            self.logger,
            self.hash,
            self._mac,
        )
        self.create_async_task(self.async_connection_setup())

    async def async_connection_setup(self):
        await self._conn.connect()
        # enable notifications
        await self._conn.write_gatt_char(0x0038 - 1, b"\x01\x00")
        # enable lower power mode
        await self._conn.write_gatt_char(0x0046 - 1, b"\xf4\x01\x00")

    def received_notification(self, uuid, data):
        self.create_async_task(self.update_data(data))

    async def update_data(self, data):
        temp = int.from_bytes(data[0:2], byteorder="little", signed=True) / 100
        humidity = int.from_bytes(data[2:3], byteorder="little")
        voltage = int.from_bytes(data[3:5], byteorder="little") / 1000.0
        batteryLevel = min(
            int(round((voltage - 2.1), 2) * 100), 100
        )  # 3.1 or above --> 100% 2.1 --> 0 %
        if batteryLevel < 10:
            low_bat = "low"
        else:
            low_bat = "ok"

        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdate(self.hash, "temperature", temp)
        await fhem.readingsBulkUpdate(self.hash, "humidity", humidity)
        await fhem.readingsBulkUpdate(self.hash, "voltage", voltage)
        await fhem.readingsBulkUpdate(self.hash, "batteryLevel", batteryLevel)
        await fhem.readingsBulkUpdate(self.hash, "battery", low_bat)
        await fhem.readingsEndUpdate(self.hash, 1)



"""Smartmeter Errors."""
import logging

logger = logging.getLogger(__name__)


class SmartmeterError(Exception):
    """Generic Error for Smartmeter."""

    def __init__(self, msg, code=None, error_response=""):
        """Creates a Smartmeter error with msg, code and error_response."""
        self.code = code or 0
        self.error_response = error_response
        super().__init__(msg)

    @property
    def msg(self):
        """Return msg."""
        return self.args[0]


class SmartmeterLoginError(SmartmeterError):
    """Raised when login fails."""


class SmartmeterConnectionError(SmartmeterError):
    """Raised due to network connectivity-related issues."""



import asyncio

from fhempy.lib.rct_power.api import RctPowerApiClient, ValidApiResponse
from rctclient.registry import REGISTRY

from .. import fhem, generic


class rct_power(generic.FhemModule):

    DEFAULT_OBJECTS = [
        "inverter_sn",
        "db.temp1",
        "battery.soc",
        "battery.soc_target",
        "battery.cycles",
        "battery.soh",
        "power_mng.battery_power_extern"
        "power_mng.soc_max",
        "power_mng.soc_min",
        "power_mng.soc_strategy"
        "battery.temperature",
        "battery.efficiency",
        "g_sync.p_acc_lp",
        "g_sync.p_ac_grid_sum_lp",
        "g_sync.p_ac_sum_lp",
        "energy.e_ac_day",
        "energy.e_grid_feed_day",
        "energy.e_load_day",
        "energy.e_grid_load_day",
    ]

    def __init__(self, logger):
        super().__init__(logger)

        self.update_loop_task = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        attr_config = {
            "interval": {
                "default": 10,
                "format": "int",
                "help": "Poll interval in seconds, default is 10s.",
            },
            "disable": {
                "default": 0,
                "format": "int",
                "options": "0,1",
            },
            "update_readings": {
                "default": "on_change",
                "options": "always,on_change",
            },
            "default_device_readings": {
                "default": "on",
                "help": "When off only readings defined in device_readings and device_readings_json are updated.",
            },
            "error_reading": {
                "default": "on",
                "help": "Use separate error reading.",
            },
            "device_readings": {
                "default": "",
                "format": "array",
                "help": (
                    "Add further objects/readings which should be retrieved<br>"
                    + "Find details about possible objects here:<br>"
                    + '<a href="https://rctclient.readthedocs.io/en/latest/'
                    + "inverter_registry.html"
                    + '">Object Infos</a><br><br>'
                    + "Example: battery.bms_sn,battery.bms_power_version"
                ),
            },
            "device_readings_json": {
                "default": "{}",
                "format": "json",
                "options": "textField-long",
                "help": "Advanced JSON configuration<br>"
                + "Example:<br>"
                + "{<br>"
                + '&nbsp;&nbsp;&nbsp;&nbsp;"battery.soc_target": {<br>'
                + '&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"reading": "battery_soc_target",<br>'
                + '&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"factor":100<br>'
                + '&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"format":".2f"<br>'
                + "&nbsp;&nbsp;&nbsp;&nbsp;}<br>"
                + "}",
            },
        }
        await self.set_attr_config(attr_config)

        set_config = {
            "display_brightness": {
                "args": ["value"],
                "params": {"value": {"format": "int"}},
                "help": "Display brightness [-] display_struct.brightness",
                "options": "slider,0,1,255",
                "function": "set_rct_write",
                "function_param": 0x29BDA75F,
            },
            "display_contrast": {
                "args": ["value"],
                "params": {"value": {"format": "int"}},
                "help": "Display contrast [-] display_struct.contrast",
                "options": "slider,0,1,255",
                "function": "set_rct_write",
                "function_param": 0xF247BB16,
            },
            "max_compensation_power": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Maximum compensation power from battery [W].<br>Registry: p_rec_lim[0]",
                "options": "slider,0,1,10000",
                "function": "set_rct_write",
                "function_param": 0x85886E2E,
            },
            "max_power_batt2grid": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Battery power to grid limitation [W] (normally 0).<br>Registry: p_rec_lim[1]",
                "options": "slider,0,1,10000",
                "function": "set_rct_write",
                "function_param": 0x54829753,
            },
            "max_power_ac": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Maximum inverter AC power [W].<br>Registry: p_rec_lim[2]",
                "options": "slider,0,1,10000",
                "function": "set_rct_write",
                "function_param": 0x9A67600D,
            },
            "battmng_kreserve": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Reserve coefficent for battery power management [-] Higher values = lower battery SoC (higher reserve). Decimal values from 0.00 - 2.00 are allowed.<br>Registry: bat_mng_struct.k_reserve",
                "function": "set_rct_write",
                "function_param": 0xF644DCA7,
            },
            "battmng_ktrust": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Time constant for battery power management [Seconds] Lower values = faster reaction on weather prediction. Decimal values from 0.00 - 1000 are allowed.<br>Registry: bat_mng_struct.k_trust",
                "function": "set_rct_write",
                "function_param": 0xB2FB9A90,
            },
            "batt_soc_min": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Minimum battery SoC target [-]. Decimal values from 0.00 - 1.00 are allowed.<br>Registry: power_mng.soc_min",
                "function": "set_rct_write",
                "function_param": 0xCE266F0F,
            },
            "batt_soc_max": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": (
                    "Maximum battery SoC target [-]. Decimal values from 0.00 - 1.00 are allowed."
                    "<br>Registry: power_mng.soc_max"
                ),
                "function": "set_rct_write",
                "function_param": 0x97997C93,
            },
            "conv_max_current": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Maximum battery converter current (charge/discharge) [A]. Decimal values from 0.00 - 25.00 are allowed.<br>Registry: acc_conv.i_max",
                "function": "set_rct_write",
                "function_param": 0xE3F4D1DF,
            },
            "conv_max_dischargecurrent": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Maximum battery discharge current [A]. Decimal values from 0.00 - 25.00 are allowed.<br>Registry: acc_conv.i_discharge_max",
                "function": "set_rct_write",
                "function_param": 0xC642B9D6,
            },
            "conv_max_chargecurrent": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Maximum battery charge current [A]. Decimal values from 0.00 - 25.00 are allowed.<br>Registry: acc_conv.i_charge_max",
                "function": "set_rct_write",
                "function_param": 0xB0FA4D23,
            },
            "max_power_ac": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Maximum inverter AC power [W]. Decimal values from 0 - 10000 are allowed.<br>Registry: p_rec_lim[2]",
                "function": "set_rct_write",
                "function_param": 0x9A67600D,
            },
            "min_soc_maint_charge": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Min SoC for start maintenance charging [-]. Decimal values from 0.00 - 1.00 are allowed.<br>Registry: power_mng.soc_charge",
                "function": "set_rct_write",
                "function_param": 0xBD3A23C3,
            },
            "plant_power_reduction": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Regulatory power limitation of plant. Germany=0.7 [-]. Decimal values from 0.00 - 1.00 are allowed.<br>Registry: buf_v_control.power_reduction",
                "function": "set_rct_write",
                "function_param": 0xFE1AA500,
            },
            "max_power_batt2grid": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Battery power to grid limitation [W] (normally 0).<br>Registry: p_rec_lim[1]",
                "options": "slider,0,1,10000",
                "function": "set_rct_write",
                "function_param": 0x54829753,
            },
            "power_mng.battery_power_extern": {
                "args": ["value"],
                "params": {"value": {"format": "float"}},
                "help": "Battery Power [W] Battery target power (positive = discharge)<br>Registry: power_mng.battery_power_extern<br>Default: 0",
                "function": "set_rct_write",
                "function_param": 0xBD008E29,
            },
            "power_mng.soc_strategy": {
                "args": ["value"],
                "params": {"value": {"format": "int"}},
                "help": "Battery SoC strategy [-] 0: 'SOC target = SOC', 1: 'Constant', 2: 'External', 3: 'Middle battery voltage', 4: 'Internal', 5: 'Schedule'<br>Registry: power_mng.soc_strategy<br>Default: 4",
                "function": "set_rct_write",
                "function_param": 0xF168B748,
            },
        }
        await self.set_set_config(set_config)
        
        if len(args) < 4 or len(args) > 5:
            return "Usage: define my_rct fhempy rct_power IP [PORT]"

        self._hostname = args[3]
        self._port = 8899
        if len(args) == 5:
            self._port = args[4]

        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "connecting")
        await fhem.readingsEndUpdate(hash, 1)

        self.create_async_task(self.setup_rct())

    async def set_rct_write(self, hash, params):
        self.create_async_task(
            self.rctclient.async_send_data(params["function_param"], params["value"])
        )

    async def set_attr_disable(self, hash):
        if self._attr_disable == 1:
            self.cancel_async_task(self.update_loop_task)
            self.update_loop_task = None
        else:
            if self.update_loop_task is not None:
                self.update_loop_task = self.create_async_task(self.update_loop())

    async def setup_rct(self):
        self.rctclient = RctPowerApiClient(
            self.logger, hostname=self._hostname, port=self._port
        )
        self.update_loop_task = self.create_async_task(self.update_loop())

    async def update_loop(self):
        while True:
            if self._attr_disable == 1:
                return

            try:
                await self.update_readings()
            except Exception:
                self.logger.exception("Failed update_readings in loop")
            await asyncio.sleep(self._attr_interval)

    async def update_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        try:
            object_ids = []
            retrieve_objects = [
                *self._attr_device_readings,
                *list(self._attr_device_readings_json),
            ]
            if self._attr_default_device_readings == "on":
                retrieve_objects = [
                    *rct_power.DEFAULT_OBJECTS,
                    *retrieve_objects,
                ]
            for val in retrieve_objects:
                for object_info in REGISTRY.all():
                    if object_info.name == val:
                        object_ids.append(object_info.object_id)

            response = await self.rctclient.async_get_data(object_ids)
            for object_id in response:
                if self._attr_error_reading == "on":
                    await fhem.readingsBulkUpdateIfChanged(
                        self.hash,
                        "error",
                        "",
                    )
                # set reading name from attribute config
                reading = response[object_id].object_name
                if reading in self._attr_device_readings_json:
                    reading = self._attr_device_readings_json[reading].get(
                        "reading", response[object_id].object_name
                    )

                if isinstance(response[object_id], ValidApiResponse):
                    # do factor calculation for float values
                    value = response[object_id].value
                    if isinstance(value, float):
                        factor = self._attr_device_readings_json.get(
                            response[object_id].object_name, {}
                        ).get("factor", 1)
                        value = value * factor
                        format = self._attr_device_readings_json.get(
                            response[object_id].object_name, {}
                        ).get("format", ".2f")
                        value = f"{value:{format}}"

                    await self.readingsBulkUpdate(
                        self.hash,
                        reading,
                        value,
                    )
                else:
                    if self._attr_error_reading == "on":
                        await self.readingsBulkUpdate(
                            self.hash,
                            "error",
                            f"{reading} failed with {response[object_id].cause}",
                        )
                    else:
                        await self.readingsBulkUpdate(
                            self.hash,
                            reading,
                            response[object_id].cause,
                        )

            await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "connected")

        except Exception:
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "connection error")
            self.logger.exception("Failed to update_readings")
        await fhem.readingsEndUpdate(self.hash, 1)

    async def readingsBulkUpdate(self, hash, reading, value):
        if self._attr_update_readings == "always":
            await fhem.readingsBulkUpdate(hash, reading, value)
        else:
            await fhem.readingsBulkUpdateIfChanged(hash, reading, value)



import asyncio

from .. import fhem, generic
from pymyenergi.connection import Connection
from pymyenergi.zappi import Zappi
from pymyenergi.harvi import Harvi

class zappi(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.zappi_box = None
        self.harvis = []
        self.harvi_serials = []

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        attr_config = {
            "interval": {
                "default": 60,
                "format": "int",
                "help": "Change interval, default is 60.",
            }
        }
        await self.set_attr_config(attr_config)

        set_config = {
            "charge_mode": {
                "args": ["charge_mode"],
                "argsh": ["charge_mode"],
                "params": {"charge_mode": {"default": "Stopped", "optional": False}},
                "options": "Fast,Eco,Eco+,Stopped",
            },
            "green_energy_ratio": {"args": ["green_energy"], "options": "slider,0,10,100"},
            
           
        }
        await self.set_set_config(set_config)

        if len(args) < 5:
            return "Usage: define myZappiBox fhempy zappi serialnumber API_Key [harvi1] [harvi2] ... [harviN] \n [hari1] ... [harviN] are optional harvi serials"
        
        self.serialNo = args[3]
        self.Apikey = args[4]
        
        #get harvi serials from args
        if len(args) > 5:
            self.harvi_serials = args[5:]

        self.create_async_task(self.setup_connection())

    async def setup_connection(self):
            """
            Sets up the connection to the Zappi device and initializes the Zappi and Harvi objects.
            """
                
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connecting",1)

            conn = Connection(self.serialNo, self.Apikey)
            self.zappi_box = Zappi(conn, self.serialNo)
           
            #iterate through harvi serials and create harvi objects
            for serial in self.harvi_serials:
                self.harvis.append(Harvi(conn, serial)) 

            if self.zappi_box != None:
                await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connected",1)
                await self.zappi_box.refresh()
                            
                await fhem.readingsSingleUpdateIfChanged(self.hash, "zappi_serial", self.zappi_box.serial_number, 1)
                await fhem.readingsSingleUpdateIfChanged(self.hash, "zappi_firmware_version", self.zappi_box.firmware_version, 1)

                # add harvi serial numbers for every single harvi 
                for harvi in self.harvis:
                    await harvi.refresh()
                    harviname = ''.join(e for e in harvi.ct1.name if e.isalnum())
                    await fhem.readingsSingleUpdateIfChanged(self.hash, "harvi_serial_" + harviname, harvi.serial_number, 1)
                    await fhem.readingsSingleUpdateIfChanged(self.hash, "harvi_firmware_" + harviname, harvi.firmware_version, 1)
                
                await self.update_zappi_data()
            else:
                await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "disconnected")

      

    async def set_green_energy_ratio(self, hash, params):
        # update green energy ratio
        green_eng = params["green_energy"]
        await fhem.readingsSingleUpdate(hash, "green_energy_min", green_eng, 1)
        await self.zappi_box.set_minimum_green_level(green_eng)
        
    async def long_running_task(self):
        await asyncio.sleep(30)
        await fhem.readingsSingleUpdate(self.hash, "state", "long running off", 1)

    async def set_charge_mode(self, hash, params):
        # set charge mode
        mode = params["charge_mode"]
        #self.zappi_box.refresh
        await self.zappi_box.set_charge_mode(mode)
        await fhem.readingsSingleUpdate(self.hash, "charge_mode", mode, 1)

   
        

    async def update_zappi_data(self):
            """
            Updates the readings of the Zappi device and its associated Harvi devices.
            """
            while True:
                
                try:
                    await self.zappi_box.refresh()
                    
                    # Update Zappi readings
                    await fhem.readingsBeginUpdate(self.hash)
                    await fhem.readingsBulkUpdate(self.hash, "operating_status", self.zappi_box.status)
                    await fhem.readingsBulkUpdate(self.hash, "plug_status", self.zappi_box.plug_status)
                    await fhem.readingsBulkUpdate(self.hash, "green_energy_min", self.zappi_box.minimum_green_level)
                    await fhem.readingsBulkUpdate(self.hash, "green_energy_total", self.zappi_box.energy_green)
                    await fhem.readingsBulkUpdate(self.hash, "energy_total", self.zappi_box.energy_total)
                    await fhem.readingsBulkUpdate(self.hash, "energy_today", self.zappi_box.energy_total)
                    await fhem.readingsBulkUpdate(self.hash, "firmware_update_available", self.zappi_box.update_available)
                    await fhem.readingsBulkUpdate(self.hash, "charge_mode", self.zappi_box.charge_mode)
                    await fhem.readingsBulkUpdate(self.hash, "locked", self.zappi_box.locked)
                    await fhem.readingsBulkUpdate(self.hash, "lock_when_plugged", self.zappi_box.lock_when_pluggedin)
                    await fhem.readingsBulkUpdate(self.hash, "lock_when_unplugged", self.zappi_box.lock_when_unplugged)
                    await fhem.readingsBulkUpdate(self.hash, "charge_when_locked", self.zappi_box.charge_when_locked)
                    await fhem.readingsBulkUpdate(self.hash, "charge_session_allowed", self.zappi_box.charge_session_allowed)
                    await fhem.readingsBulkUpdate(self.hash, "power_grid", self.zappi_box.power_grid)
                    await fhem.readingsBulkUpdate(self.hash, "power_generated", self.zappi_box.power_generated)
                    await fhem.readingsBulkUpdate(self.hash, "smart_boosted_energy", self.zappi_box.boost_amount)
                    await fhem.readingsBulkUpdate(self.hash, "power_ct1", self.zappi_box.ct1.power)
                    await fhem.readingsBulkUpdate(self.hash, "power_ct2", self.zappi_box.ct2.power)
                    await fhem.readingsBulkUpdate(self.hash, "power_ct3", self.zappi_box.ct3.power)
                    await fhem.readingsBulkUpdate(self.hash, "power_total", self.zappi_box.ct1.power + self.zappi_box.ct2.power + self.zappi_box.ct3.power)

                    # Iterate over the Harvi devices and update readings
                    for harvi in self.harvis:
                        await harvi.refresh()
                        harviname = ''.join(e for e in harvi.ct1.name if e.isalnum())
                        await fhem.readingsBulkUpdate(self.hash, harviname + "_CT1", harvi.ct1.power)
                        await fhem.readingsBulkUpdate(self.hash, harviname + "_CT2", harvi.ct2.power)
                        await fhem.readingsBulkUpdate(self.hash, harviname + "_CT3", harvi.ct3.power)
                        
                       
                    """
                    if self.zappi_box.smart_boost_start_hour > -1:
                        smartBoostStart = self.zappi_box.smart_boost_start_hour + ":" + self.zappi_box.smart_boost_start_minute
                        await fhem.readingsBulkUpdate(self.hash, "smart_boost_start",  smartBoostStart)
                    else:
                        await fhem.readingsBulkUpdate(self.hash, "smart_boost_start",  "not set")
                    """
                    await fhem.readingsBulkUpdate(self.hash, "smart_boost_amount", self.zappi_box.smart_boost_amount)
                    await fhem.readingsEndUpdate(self.hash, 1)
                except Exception:
                    self.logger.error("Failed to update zappi readings")
                    
                
                
                await asyncio.sleep(self._attr_interval)

    


import asyncio
import datetime

from tibber import Tibber

from .. import fhem, generic


class tibber(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.tibber_connection = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "interval": {
                "default": 3600,
                "format": "int",
                "help": "Change interval, default is 3600s.",
            }
        }
        await self.set_attr_config(attr_config)

        if len(args) != 4:
            return "Usage: define mytibber fhempy tibber TOKEN"

        self.token = args[3]

        self.create_async_task(self.setup_connection())

    async def setup_connection(self):
        self.tibber_connection = Tibber(self.token, user_agent="fhempy")
        await self.tibber_connection.update_info()
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "tibber_name", self.tibber_connection.name, 1
        )
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "tibber_userid", self.tibber_connection.user_id, 1
        )
        await self.update_home_data()

    def _rt_callback(self, pkg):
        data = pkg.get("data")
        if data is None:
            return
        self.create_async_task(self.update_rt_data(data.get("liveMeasurement")))

    async def update_rt_data(self, data):
        await fhem.readingsBeginUpdate(self.hash)
        try:
            for name in data:
                await fhem.readingsBulkUpdate(self.hash, "rt_" + name, data[name])
        except Exception:
            self.logger.error("Failed to update realtime readings")
        
        # additional total consumption / production reading
        await fhem.readingsBulkUpdate(self.hash, "rt_powerNet" , data["power"]-data["powerProduction"])
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_home_data(self):
        home = self.tibber_connection.get_homes()[0]
        await home.update_info()

        if home.has_real_time_consumption:
            await home.rt_subscribe(self._rt_callback)

        self.create_async_task(self.update_current_price_data())
        self.create_async_task(self.update_prices_forPeriods())

        while True:
            try:
                await home.fetch_consumption_data()
                await home.update_info()
                await home.update_price_info()
            except Exception:
                self.logger.error("Failed to update data from tibber, retry in 60s")
                await asyncio.sleep(60)
                continue

            await fhem.readingsBeginUpdate(self.hash)
            try:
                for i, (timestamp, value) in enumerate(home._price_info.items()):
                    # Extracting date and time
                    date, time = timestamp.split('T')
                    time = time.split('+')[0]  # Removing timezone offset
                    hour = time.split(':')[0].lstrip('0')  # Removing leading zero from hour

                    # Determining whether it's today or tomorrow
                    day_key = 'today' if i < 24 else 'tomorrow'

                    # Creating the new key in the desired format
                    new_key = f'{day_key}_{hour.zfill(2)}00'

                    # Assigning the value to the new key in the new dictionary
                    await fhem.readingsBulkUpdate(self.hash, new_key, value)

                await fhem.readingsBulkUpdate(self.hash, "address", home.address1)
                await fhem.readingsBulkUpdate(
                    self.hash, "has_active_subscription", home.has_active_subscription
                )
                await fhem.readingsBulkUpdate(
                    self.hash, "has_production", home.has_production
                )
                await fhem.readingsBulkUpdate(
                    self.hash,
                    "has_real_time_consumption",
                    home.has_real_time_consumption,
                )
                await fhem.readingsBulkUpdate(self.hash, "home_id", home.home_id)
                await fhem.readingsBulkUpdate(self.hash, "month_cons", home.month_cons)
                await fhem.readingsBulkUpdate(self.hash, "month_cost", home.month_cost)
                await fhem.readingsBulkUpdate(self.hash, "home_name", home.name)
                await fhem.readingsBulkUpdate(self.hash, "peak_hour", home.peak_hour)

                
            except Exception:
                self.logger.error("Failed to update readings")
            await fhem.readingsEndUpdate(self.hash, 1)
            
            await asyncio.sleep(self._attr_interval)

    async def update_current_price_data(self):
            """
            Updates the current price data from Tibber API and updates the readings in FHEM.

            This method retrieves the current price information from the Tibber API and updates the corresponding readings
            in FHEM. It continuously fetches the latest price data and updates the readings every new hour.

            Raises:
                Exception: If there is an error updating the data from Tibber API or updating the readings in FHEM.

            """
            home = self.tibber_connection.get_homes()[0]
            
            while True:
                try:
                    await home.update_info()
                    await home.update_price_info()
                except Exception:
                    self.logger.error("Failed to update data from tibber, retry in 60s")
                    await asyncio.sleep(60)
                    continue

                await fhem.readingsBeginUpdate(self.hash)
                try:
                    #  price information incl price rank
                    price, level, time, rank = home.current_price_data()
                   
                    
                    
                    # price info
                    await fhem.readingsBulkUpdate(
                        self.hash, "current_price_energy", home.current_price_info["energy"]
                    )
                    await fhem.readingsBulkUpdate(
                        self.hash, "current_price_tax", home.current_price_info["tax"]
                    )
                    await fhem.readingsBulkUpdate(
                        self.hash, "current_price_total", home.current_price_info["total"]
                    )
                    await fhem.readingsBulkUpdate(
                        self.hash,
                        "current_price_startsat",
                        home.current_price_info["startsAt"],
                    )
                    await fhem.readingsBulkUpdate(
                        self.hash, "current_price_level", home.current_price_info["level"]
                    )
                    await fhem.readingsBulkUpdate(
                        self.hash, "current_price_rank", rank
                    )

                
                except Exception:
                    self.logger.error("Failed to update readings")
                await fhem.readingsEndUpdate(self.hash, 1)
                
                #  update readings every new hour to fetch new current_* data
                now = datetime.datetime.now()
                remaining_seconds = 3600 - (now.minute * 60 + now.second)
                await asyncio.sleep(remaining_seconds+5)

    async def Undefine(self, hash):
        if self.tibber_connection:
            await self.tibber_connection.close_connection()
        return await super().Undefine(hash)
    
    async def update_prices_forPeriods(self):
            """
            Updates the price information for different periods and stores the results in FHEM readings.

            This method retrieves the price information for different periods from the Tibber API and calculates
            the periods with the lowest and highest prices. It then updates the corresponding readings in FHEM.

            Raises:
                Exception: If there is an error updating the data from Tibber or updating the readings in FHEM.

            """
            home = self.tibber_connection.get_homes()[0]
            
            while True:
                try:
                    await home.update_info()
                    await home.update_price_info()
                except Exception:
                    self.logger.error("Failed to update data from tibber, retry in 60s")
                    await asyncio.sleep(60)
                    continue

                await fhem.readingsBeginUpdate(self.hash)
                try:

                    time_list = list(home._price_info.keys())
                    price_list = list(home._price_info.values())
                    now = datetime.datetime.now()

                    # find the index of the element with lowest price today
                    cheapest_price_idx_today = price_list.index(min(price_list[now.hour:24]))    

                    # find the index of the element with lowest price 
                    cheapest_price_idx_total = price_list.index(min(price_list[now.hour:]))

                    # find 3h window with lowest price
                    cheapest_price_3h_idx = min(range(now.hour, len(price_list) - 3), key=lambda i: sum(price_list[i:i+3]))
                    

                    #find 2h window with lowest price
                    cheapest_price_2h_idx = min(range(now.hour, len(price_list) - 2), key=lambda i: sum(price_list[i:i+2]))
                
                    # find the index of the element with highest price today
                    highest_price_idx_today = price_list.index(max(price_list[now.hour:24])) 

                    # find the index of the element with highest price 
                    highest_price_idx_total = price_list.index(max(price_list[now.hour:]))    

                    # find 3h window with highest price
                    highest_price_3h_idx = max(range(now.hour, len(price_list) - 3), key=lambda i: sum(price_list[i:i+3]))

                    #find 2h window with highest price
                    highest_price_2h = max(range(now.hour, len(price_list) - 2), key=lambda i: sum(price_list[i:i+2]))

                    # update readings

                    await fhem.readingsBulkUpdate(self.hash, "cheapest_price_today", price_list[cheapest_price_idx_today])
                    await fhem.readingsBulkUpdate(self.hash, "cheapest_price_total", price_list[cheapest_price_idx_total])
                    await fhem.readingsBulkUpdate(self.hash, "cheapest_price_today_startsAt", datetime.datetime.fromisoformat(time_list[cheapest_price_idx_today]))  
                    await fhem.readingsBulkUpdate(self.hash, "cheapest_price_total_startsAt", datetime.datetime.fromisoformat(time_list[cheapest_price_idx_total]))
                    await fhem.readingsBulkUpdate(self.hash, "cheapest_price_3h_window_startsAt", datetime.datetime.fromisoformat(time_list[cheapest_price_3h_idx]))
                    await fhem.readingsBulkUpdate(self.hash, "cheapest_price_2h_window_startsAt", datetime.datetime.fromisoformat(time_list[cheapest_price_2h_idx]))
                    await fhem.readingsBulkUpdate(self.hash, "highest_price_today", price_list[highest_price_idx_today])
                    await fhem.readingsBulkUpdate(self.hash, "highest_price_total", price_list[highest_price_idx_total])    
                    await fhem.readingsBulkUpdate(self.hash, "highest_price_today_startsAt", datetime.datetime.fromisoformat(time_list[highest_price_idx_today]))
                    await fhem.readingsBulkUpdate(self.hash, "highest_price_total_startsAt", datetime.datetime.fromisoformat(time_list[highest_price_idx_total]))
                    await fhem.readingsBulkUpdate(self.hash, "highest_price_3h_window_startsAt", datetime.datetime.fromisoformat(time_list[highest_price_3h_idx]))
                    await fhem.readingsBulkUpdate(self.hash, "highest_price_2h_window_startsAt", datetime.datetime.fromisoformat(time_list[highest_price_2h]))


                
                except Exception:
                    self.logger.error("Failed to update readings")
                
                await fhem.readingsEndUpdate(self.hash, 1)
                # get remaining seconds until next full hour 
                remaining_seconds = 3600 - (now.minute * 60 + now.second)
                await asyncio.sleep(remaining_seconds+5)

